# pylint: disable=missing-docstring

import logging
import os
import subprocess
import sys
import textwrap

import nengo
import numpy as np
import pytest
import tensorflow as tf
from nengo.builder.operator import Reset
from nengo.builder.signal import Signal
from nengo.exceptions import BuildError
from packaging import version

from nengo_dl import config, graph_optimizer, tensor_graph, utils
from nengo_dl.tests import dummies


@pytest.mark.parametrize("unroll", (1, 2))
@pytest.mark.training
def test_gradients(Simulator, unroll, seed):
    minibatch_size = 4

    with nengo.Network(seed=seed) as net:
        net.config[nengo.Ensemble].gain = nengo.dists.Choice([1])
        net.config[nengo.Ensemble].bias = nengo.dists.Uniform(-1, 1)

        inp = nengo.Node([0], label="inp")

        # sigmoid neurons
        ens = nengo.Ensemble(10, 1, neuron_type=nengo.Sigmoid())

        # normal decoded connection
        nengo.Connection(inp, ens)

        # recurrent connection
        nengo.Connection(ens, ens, transform=0.1)

        # rectified neurons
        ens2 = nengo.Ensemble(10, 2, neuron_type=nengo.RectifiedLinear())

        # neuron--neuron connection
        nengo.Connection(
            ens, ens2, transform=[[1], [1]], solver=nengo.solvers.LstsqL2(weights=True)
        )

        # sliced output, no synapse
        nengo.Connection(inp, ens2[0], synapse=None, transform=0.5)

        # sliced input, sliced output
        inp2 = nengo.Node([0, 0], label="inp2")
        nengo.Connection(inp2[0], ens2[1])

        nengo.Probe(ens)
        nengo.Probe(ens2)

    with Simulator(net, unroll_simulation=unroll, minibatch_size=minibatch_size) as sim:
        sim.check_gradients(atol=1e-4)


def test_mark_signals():
    with nengo.Network() as net:
        ens0 = nengo.Ensemble(10, 1, neuron_type=nengo.LIF())
        ens1 = nengo.Ensemble(20, 1, neuron_type=nengo.Direct())
        ens2 = nengo.Ensemble(30, 1)
        conn0 = nengo.Connection(ens0, ens1)
        conn1 = nengo.Connection(ens0, ens1, learning_rule_type=nengo.PES())
        conn2 = nengo.Connection(ens0, ens2, learning_rule_type=nengo.Voja())
        nengo.Probe(ens2)

    model = nengo.builder.Model()
    model.build(net)

    tg = tensor_graph.TensorGraph(
        model, None, None, 1, None, utils.NullProgressBar(), None
    )
    tg.mark_signals()

    assert model.sig[ens0]["encoders"].trainable
    assert model.sig[ens1]["encoders"].trainable
    assert not model.sig[ens2]["encoders"].trainable
    assert model.sig[ens0.neurons]["bias"].trainable
    assert model.sig[ens2.neurons]["bias"].trainable
    assert model.sig[conn0]["weights"].trainable
    assert not model.sig[conn1]["weights"].trainable
    assert model.sig[conn2]["weights"].trainable

    trainables = (
        model.sig[ens0]["encoders"],
        model.sig[ens1]["encoders"],
        model.sig[ens0.neurons]["bias"],
        model.sig[ens2.neurons]["bias"],
        model.sig[conn0]["weights"],
        model.sig[conn2]["weights"],
    )

    for op in model.operators:
        for sig in op.all_signals:
            if sig in trainables:
                assert sig.trainable
            else:
                assert not sig.trainable


def test_mark_signals_config():
    with nengo.Network() as net:
        config.configure_settings(trainable=None)
        net.config[nengo.Ensemble].trainable = False

        with nengo.Network():
            # check that object in subnetwork inherits config from parent
            ens0 = nengo.Ensemble(10, 1, label="ens0")

            # check that ens.neurons can be set independent of ens
            net.config[ens0.neurons].trainable = True

            with nengo.Network():
                with nengo.Network():
                    # check that subnetworks can override parent configs
                    config.configure_settings(trainable=True)
                    ens1 = nengo.Ensemble(10, 1, label="ens1")

                    with nengo.Network():
                        # check that subnetworks inherit the trainable settings
                        # from parent networks
                        ens3 = nengo.Ensemble(10, 1, label="ens3")

            # check that instances can be set independent of class
            ens2 = nengo.Ensemble(10, 1, label="ens2")
            net.config[ens2].trainable = True

    model = nengo.builder.Model()
    model.build(net)

    progress = utils.NullProgressBar()

    tg = tensor_graph.TensorGraph(model, None, None, 1, None, progress, None)
    tg.mark_signals()

    assert not model.sig[ens0]["encoders"].trainable
    assert model.sig[ens0.neurons]["bias"].trainable

    assert model.sig[ens1]["encoders"].trainable

    assert model.sig[ens2]["encoders"].trainable

    assert model.sig[ens3]["encoders"].trainable

    # check that learning rule connections can be manually set to True
    with nengo.Network() as net:
        config.configure_settings(trainable=None)

        a = nengo.Ensemble(10, 1)
        b = nengo.Ensemble(10, 1)
        conn0 = nengo.Connection(a, b, learning_rule_type=nengo.PES())
        net.config[conn0].trainable = True

    model = nengo.builder.Model()
    model.build(net)

    tg = tensor_graph.TensorGraph(model, None, None, 1, None, progress, None)
    with pytest.warns(UserWarning):
        tg.mark_signals()

    assert model.sig[conn0]["weights"].trainable

    with nengo.Network() as net:
        config.configure_settings(trainable=None)

        a = nengo.Node([0])
        ens = nengo.Ensemble(10, 1)
        nengo.Connection(a, ens, learning_rule_type=nengo.Voja())
        net.config[nengo.Ensemble].trainable = True

    model = nengo.builder.Model()
    model.build(net)

    tg = tensor_graph.TensorGraph(model, None, None, 1, None, progress, None)
    with pytest.warns(UserWarning):
        tg.mark_signals()

    assert model.sig[ens]["encoders"].trainable

    # check that models with no toplevel work
    sig = nengo.builder.signal.Signal([0])
    op = nengo.builder.operator.Reset(sig, 1)
    model = nengo.builder.Model()
    model.add_op(op)

    tg = tensor_graph.TensorGraph(model, None, None, 1, None, progress, None)
    with pytest.warns(UserWarning):
        tg.mark_signals()

    assert not sig.trainable


@pytest.mark.parametrize("config_planner", (True, False, None))
def test_planner_config(config_planner):
    with nengo.Network() as net:
        if config_planner is not None:
            net.config.configures(nengo.Network)
            if config_planner:
                net.config[nengo.Network].set_param(
                    "planner",
                    nengo.params.Parameter("planner", graph_optimizer.noop_planner),
                )

    model = nengo.builder.Model()
    model.build(net)
    sig = nengo.builder.signal.Signal([1])
    sig2 = nengo.builder.signal.Signal([1])
    sig3 = nengo.builder.signal.Signal([1])
    model.add_op(nengo.builder.operator.DotInc(sig, sig2, sig3))
    model.add_op(nengo.builder.operator.DotInc(sig, sig2, sig3))

    tg = tensor_graph.TensorGraph(
        model, None, None, 1, None, utils.NullProgressBar(), None
    )

    assert len(tg.plan) == (3 if config_planner else 2)


def test_signal_order_deterministic(Simulator, seed):
    with nengo.Network(seed=seed) as net:
        inp = nengo.Node([0])
        ens0 = nengo.Ensemble(10, 1, label="ens")
        ens1 = nengo.Ensemble(10, 1, label="ens")
        nengo.Connection(inp, ens0, synapse=None)
        nengo.Connection(inp, ens1, synapse=None)

    with Simulator(net, seed=seed) as sim1:
        with Simulator(net, seed=seed) as sim2:
            for trainable in ("trainable", "non_trainable", "state"):
                for v, v2 in zip(
                    sim1.tensor_graph.base_arrays_init[trainable].values(),
                    sim2.tensor_graph.base_arrays_init[trainable].values(),
                ):
                    assert all(
                        (x is None and y is None) or np.allclose(x, y)
                        for x, y in zip(v[0], v2[0])
                    )


def test_create_signals():
    # check that floats/ints get split into different arrays

    sigs = [
        dummies.Signal(dtype=np.float32),
        dummies.Signal(dtype=np.float32),
        dummies.Signal(dtype=np.int32),
        dummies.Signal(dtype=np.int32),
    ]
    plan = [tuple(dummies.Op(reads=[x]) for x in sigs)]
    graph = dummies.TensorGraph(plan, tf.float32, 10)
    graph.create_signals(sigs)

    assert graph.signals[sigs[0]].key == graph.signals[sigs[1]].key
    assert graph.signals[sigs[1]].key != graph.signals[sigs[2]].key
    assert graph.signals[sigs[2]].key == graph.signals[sigs[3]].key

    # check that floats all get converted to same precision and combined
    sigs = [
        dummies.Signal(dtype=np.float32),
        dummies.Signal(dtype=np.float32),
        dummies.Signal(dtype=np.float64),
        dummies.Signal(dtype=np.float64),
    ]
    plan = [tuple(dummies.Op(reads=[x]) for x in sigs)]
    graph = dummies.TensorGraph(plan, tf.float32, 10)
    graph.create_signals(sigs)
    assert np.all([graph.signals[x].dtype == "float32" for x in sigs])
    assert graph.signals[sigs[0]].key == graph.signals[sigs[1]].key
    assert graph.signals[sigs[1]].key == graph.signals[sigs[2]].key
    assert graph.signals[sigs[2]].key == graph.signals[sigs[3]].key

    # check that ints all get converted to same precision and combined
    sigs = [
        dummies.Signal(dtype=np.int32),
        dummies.Signal(dtype=np.int32),
        dummies.Signal(dtype=np.int64),
        dummies.Signal(dtype=np.int64),
    ]
    plan = [tuple(dummies.Op(reads=[x]) for x in sigs)]
    graph = dummies.TensorGraph(plan, tf.float32, 10)
    graph.create_signals(sigs)
    assert np.all([graph.signals[x].dtype == "int32" for x in sigs])
    assert graph.signals[sigs[0]].key == graph.signals[sigs[1]].key
    assert graph.signals[sigs[1]].key == graph.signals[sigs[2]].key
    assert graph.signals[sigs[2]].key == graph.signals[sigs[3]].key

    # check that different shapes go in different groups
    sigs = [
        dummies.Signal(shape=(10,)),
        dummies.Signal(shape=(5,)),
        dummies.Signal(shape=(10, 1)),
        dummies.Signal(shape=(5, 1)),
    ]
    plan = [tuple(dummies.Op(reads=[x]) for x in sigs)]
    graph = dummies.TensorGraph(plan, tf.float32, 10)
    graph.create_signals(sigs)
    assert graph.base_arrays_init["non_trainable"][graph.signals[sigs[0]].key][1] == [
        (10, 10),
        (10, 5),
    ]
    assert graph.base_arrays_init["non_trainable"][graph.signals[sigs[2]].key][1] == [
        (10, 10, 1),
        (10, 5, 1),
    ]
    assert graph.signals[sigs[0]].key == graph.signals[sigs[1]].key
    assert graph.signals[sigs[1]].key != graph.signals[sigs[2]].key
    assert graph.signals[sigs[2]].key == graph.signals[sigs[3]].key

    # check trainable
    sigs = [
        dummies.Signal(trainable=True),
        dummies.Signal(trainable=True),
        dummies.Signal(trainable=False),
        dummies.Signal(trainable=False),
    ]
    plan = [tuple(dummies.Op(reads=[x]) for x in sigs)]
    graph = dummies.TensorGraph(plan, tf.float32, 10)
    graph.create_signals(sigs)
    assert graph.base_arrays_init["trainable"][graph.signals[sigs[0]].key][1] == [
        (1,),
        (1,),
    ]
    assert graph.base_arrays_init["non_trainable"][graph.signals[sigs[2]].key][1] == [
        (10, 1),
        (10, 1),
    ]
    assert graph.signals[sigs[0]].key == graph.signals[sigs[1]].key
    assert graph.signals[sigs[1]].key != graph.signals[sigs[2]].key
    assert graph.signals[sigs[2]].key == graph.signals[sigs[3]].key

    # check that scalars get upsized
    sigs = [dummies.Signal(shape=()), dummies.Signal(shape=(4,))]
    plan = [tuple(dummies.Op(reads=[x]) for x in sigs)]
    graph = dummies.TensorGraph(plan, tf.float32, 10)
    graph.create_signals(sigs)
    assert list(graph.base_arrays_init["non_trainable"].values())[0][1] == [
        (10, 1),
        (10, 4),
    ]

    # check that boolean signals are handled correctly
    sigs = [dummies.Signal(dtype=np.bool, shape=())]
    plan = [(dummies.Op(reads=sigs),)]
    graph = dummies.TensorGraph(plan, tf.float32, 1)
    graph.create_signals(sigs)
    assert list(graph.base_arrays_init["non_trainable"].values())[0][2] == "bool"


def test_create_signals_views():
    sigs = [
        dummies.Signal(shape=(2, 2), base_shape=(4,)),
        dummies.Signal(shape=(2, 2), base_shape=(4,)),
    ]
    sigs += [sigs[0].base, sigs[1].base]
    plan = [tuple(dummies.Op(reads=[x]) for x in sigs)]
    graph = dummies.TensorGraph(plan, tf.float32, 10)
    graph.create_signals(sigs[2:])
    assert list(graph.base_arrays_init["non_trainable"].values())[0][1] == [
        (10, 4),
        (10, 4),
    ]
    assert graph.signals[sigs[0]].key == graph.signals[sigs[1]].key
    assert graph.signals[sigs[1]].key == graph.signals[sigs[2]].key
    assert graph.signals[sigs[2]].key == graph.signals[sigs[3]].key
    assert graph.signals[sigs[0]].slices == ((0, 4),)
    assert graph.signals[sigs[1]].slices == ((4, 8),)
    assert graph.signals[sigs[0]].slices == graph.signals[sigs[2]].slices
    assert graph.signals[sigs[1]].slices == graph.signals[sigs[3]].slices

    slice_sig = Signal(shape=(10, 4, 3, 2))[1:8:2]
    slice_sig.trainable = False
    slice_sig.minibatched = True
    slice_sig.base.trainable = False
    slice_sig.base.minibatched = True
    plan = [(dummies.Op(reads=[slice_sig]),)]
    graph = dummies.TensorGraph(plan, tf.float32, 10)
    graph.create_signals([slice_sig.base])
    assert graph.signals[slice_sig].key == graph.signals[slice_sig.base].key
    assert graph.signals[slice_sig].slices == ((1, 2), (3, 4), (5, 6), (7, 8))


def test_create_signals_partition():
    # check that signals are partitioned based on plan
    sigs = [dummies.Signal(), dummies.Signal(), dummies.Signal(), dummies.Signal()]
    plan = [
        tuple(dummies.Op(reads=[x]) for x in sigs[:2]),
        tuple(dummies.Op(reads=[x]) for x in sigs[2:]),
    ]
    graph = dummies.TensorGraph(plan, tf.float32, 10)
    graph.create_signals(sigs)
    assert graph.signals[sigs[0]].key == graph.signals[sigs[1]].key
    assert graph.signals[sigs[1]].key != graph.signals[sigs[2]].key
    assert graph.signals[sigs[2]].key == graph.signals[sigs[3]].key

    # check that signals are partitioned for different read blocks
    plan = [tuple(dummies.Op(reads=[sigs[i], sigs[2 + i]]) for i in range(2))]
    graph = dummies.TensorGraph(plan, tf.float32, 10)
    graph.create_signals(sigs)
    assert graph.signals[sigs[0]].key == graph.signals[sigs[1]].key
    assert graph.signals[sigs[1]].key != graph.signals[sigs[2]].key
    assert graph.signals[sigs[2]].key == graph.signals[sigs[3]].key

    # check that signals are partitioned for different sig types
    plan = [tuple(dummies.Op(reads=[sigs[i]], sets=[sigs[2 + i]]) for i in range(2))]
    graph = dummies.TensorGraph(plan, tf.float32, 10)
    graph.create_signals(sigs)
    assert graph.signals[sigs[0]].key == graph.signals[sigs[1]].key
    assert graph.signals[sigs[1]].key != graph.signals[sigs[2]].key
    assert graph.signals[sigs[2]].key == graph.signals[sigs[3]].key

    # check that resets are ignored
    sigs = [dummies.Signal(), dummies.Signal(), dummies.Signal(), dummies.Signal()]
    plan = [tuple(Reset(x) for x in sigs)]
    graph = dummies.TensorGraph(plan, tf.float32, 10)
    graph.create_signals(sigs)
    assert len(graph.base_arrays_init["non_trainable"]) == 4


@pytest.mark.parametrize("use_loop", (True, False))
def test_get_tensor(Simulator, use_loop):
    with nengo.Network() as net:
        config.configure_settings(use_loop=use_loop)

        a = nengo.Node([1])
        b = nengo.Ensemble(10, 1)
        c = nengo.Connection(
            a, b.neurons, transform=np.arange(10)[:, None], synapse=None
        )
        p = nengo.Probe(c)

        # build a signal probe so that the indices get loaded into the sim
        # (checks that the indices reloading works properly)
        nengo.Probe(c, "weights")

    kwargs = dict() if use_loop else dict(unroll_simulation=10)
    with Simulator(net, **kwargs) as sim:
        tensor = sim.tensor_graph.get_tensor(sim.model.sig[c]["weights"])

        assert np.allclose(tf.keras.backend.get_value(tensor), np.arange(10)[:, None])

        sim.run_steps(10)
        assert np.allclose(sim.data[p], np.arange(10)[None, :])


@pytest.mark.eager_only
@pytest.mark.parametrize("trainable", (True, False))
def test_build(trainable, rng):
    sigs = [
        dummies.Signal(
            shape=(2, 1), dtype="float32", initial_value=0, trainable=trainable
        ),
        dummies.Signal(
            shape=(3, 1),
            dtype="float32",
            initial_value=np.zeros((3, 1)),
            trainable=trainable,
        ),
        dummies.Signal(
            shape=(4, 1), dtype="float32", initial_value=1, trainable=trainable
        ),
        dummies.Signal(
            shape=(5, 1),
            dtype="float32",
            initial_value=np.ones((5, 1)),
            trainable=trainable,
        ),
        dummies.Signal(
            shape=(6, 1),
            dtype="float32",
            initial_value=rng.uniform(size=(6, 1)),
            trainable=trainable,
        ),
        dummies.Signal(
            shape=(7, 1),
            dtype="float32",
            initial_value=rng.uniform(size=(7, 1)),
            trainable=trainable,
        ),
    ]

    plan = [
        tuple(dummies.Op(reads=[x]) for x in sigs[:2]),
        tuple(dummies.Op(reads=[x]) for x in sigs[2:4]),
        tuple(dummies.Op(reads=[x]) for x in sigs[4:]),
    ]

    graph = dummies.TensorGraph(plan=plan, dtype="float32", minibatch_size=16)
    graph.create_signals(sigs)
    graph.build()

    if trainable:
        assert len(graph.trainable_weights) == 3
        assert len(graph.non_trainable_weights) == 0
    else:
        assert len(graph.trainable_weights) == 0
        assert len(graph.non_trainable_weights) == 3

    init0 = graph.weights[0].numpy()
    assert init0.shape == (5, 1) if trainable else (16, 5, 1)
    assert np.allclose(init0, 0)

    init1 = graph.weights[1].numpy()
    assert init1.shape == (9, 1) if trainable else (16, 9, 1)
    assert np.allclose(init1, 1)

    init2 = graph.weights[2].numpy()
    if trainable:
        assert init2.shape == (13, 1)
        assert np.allclose(init2[:6], sigs[4].initial_value)
        assert np.allclose(init2[6:], sigs[5].initial_value)
    else:
        assert init2.shape == (16, 13, 1)
        assert np.allclose(init2[:, :6], sigs[4].initial_value)
        assert np.allclose(init2[:, 6:], sigs[5].initial_value)


@pytest.mark.parametrize("use_loop", (True, False))
def test_conditional_update(Simulator, use_loop, caplog):
    caplog.set_level(logging.INFO)

    with nengo.Network() as net:
        config.configure_settings(stateful=False, use_loop=use_loop)

        a = nengo.Ensemble(10, 1)
        b = nengo.Node(size_in=1)
        conn = nengo.Connection(a, b)

    with Simulator(net):
        pass

    assert "Number of state updates: 0" in caplog.text
    caplog.clear()

    conn.learning_rule_type = nengo.PES()

    with Simulator(net):
        pass

    assert "Number of state updates: 1" in caplog.text
    caplog.clear()

    with net:
        config.configure_settings(trainable=True)

    with Simulator(net):
        pass

    assert "Number of state updates: 1" in caplog.text


def test_unsupported_op_error():
    class MyOp(dummies.Op):  # pylint: disable=abstract-method
        pass

    model = nengo.builder.Model()
    model.add_op(MyOp())
    with pytest.raises(BuildError, match="No registered builder"):
        tensor_graph.TensorGraph(model, None, None, None, None, None, None)


@pytest.mark.skipif(
    sys.version_info < (3, 6, 0), reason="order is not deterministic in python<3.6"
)
@pytest.mark.parametrize(
    "planner", ("greedy_planner", "tree_planner", "transitive_planner", "noop_planner")
)
def test_deterministic_order(planner, tmp_path):
    if version.parse(nengo.__version__) <= version.parse("3.1.0") and planner in (
        "greedy_planner",
        "noop_planner",
    ):
        # we could make this work by backporting the deterministic toposort from
        # nengo>3.1.0, but that doesn't seem worth it to get these two niche planners
        # to be deterministic
        pytest.skip(f"'{planner}' is nondeterministic in Nengo<=3.1.0")

    code = textwrap.dedent(
        f"""
    import nengo
    import nengo_dl

    with nengo.Network(seed=0) as net:
        nengo_dl.configure_settings(planner=nengo_dl.graph_optimizer.{planner})

        # use ensemblearrays as they create a lot of parallel ops
        ens0 = nengo.networks.EnsembleArray(1, 10)
        ens1 = nengo.networks.EnsembleArray(1, 10)
        nengo.Connection(ens0.output, ens1.input)
        nengo.Probe(ens1.output)

    model = nengo_dl.builder.NengoModel(
        dt=0.001,
        builder=nengo_dl.builder.NengoBuilder(),
        fail_fast=False,
    )
    model.build(net)

    tensor_graph = nengo_dl.tensor_graph.TensorGraph(
        model=model,
        dt=0.001,
        unroll_simulation=1,
        minibatch_size=1,
        device=None,
        progress=nengo_dl.utils.NullProgressBar(),
        seed=0,
    )

    plan = tensor_graph.plan
    sigs = tensor_graph.signals.sig_map

    for ops in plan:
        print(len(ops))
        for op in ops:
            print(type(op))
            for s in op.all_signals:
                print(s._name)
                print(s.shape)
                print(s.initial_value)

    for s in sigs:
        print(s._name)
        print(s.shape)
        print(s.initial_value)
    """
    )
    tmp_path = tmp_path / "test.py"
    tmp_path.write_text(code, encoding="utf-8")

    env = os.environ.copy()

    env["PYTHONHASHSEED"] = "0"
    output0 = subprocess.run(
        [sys.executable, str(tmp_path)],
        stdout=subprocess.PIPE,
        env=env,
        encoding="utf-8",
        check=True,
    )

    env["PYTHONHASHSEED"] = "1"
    output1 = subprocess.run(
        [sys.executable, str(tmp_path)],
        stdout=subprocess.PIPE,
        env=env,
        encoding="utf-8",
        check=True,
    )

    assert output0.stdout == output1.stdout



"""
Build classes for Nengo process operators.
"""

import contextlib
import logging

import numpy as np
import tensorflow as tf
from nengo.builder.processes import SimProcess
from nengo.exceptions import SimulationError
from nengo.synapses import LinearFilter, Lowpass
from nengo.utils.filter_design import cont2discrete

from nengo_dl import compat, utils
from nengo_dl.builder import Builder, OpBuilder

logger = logging.getLogger(__name__)


class GenericProcessBuilder(OpBuilder):
    """
    Builds all process types for which there is no custom TensorFlow
    implementation.

    Notes
    -----
    These will be executed as native Python functions, requiring execution to
    move in and out of TensorFlow.  This can significantly slow down the
    simulation, so any performance-critical processes should consider
    adding a custom TensorFlow implementation for their type instead.
    """

    def build_pre(self, signals, config):
        super().build_pre(signals, config)

        self.time_data = signals[self.ops[0].t].reshape(())
        self.input_data = (
            None
            if self.ops[0].input is None
            else signals.combine([op.input for op in self.ops])
        )
        self.output_data = signals.combine([op.output for op in self.ops])
        self.state_data = [
            signals.combine([list(op.state.values())[i] for op in self.ops])
            for i in range(len(self.ops[0].state))
        ]
        self.mode = "inc" if self.ops[0].mode == "inc" else "update"

        self.prev_result = []

        # `merged_func` calls the step function for each process and
        # combines the result
        def merged_func(time, *input_state):  # pragma: no cover (runs in TF)
            if not hasattr(self, "step_fs"):
                raise SimulationError(f"build_post has not been called for {self}")

            if self.input_data is None:
                input = None
                state = input_state
            else:
                input = input_state[0]
                state = input_state[1:]

            # update state in-place (this will update the state values
            # inside step_fs)
            for i, s in enumerate(state):
                self.step_states[i][...] = s

            input_offset = 0
            func_output = []
            for i, op in enumerate(self.ops):
                if op.input is not None:
                    input_shape = op.input.shape[0]
                    func_input = input[:, input_offset : input_offset + input_shape]
                    input_offset += input_shape

                mini_out = []
                for j in range(signals.minibatch_size):
                    x = [] if op.input is None else [func_input[j]]
                    mini_out += [self.step_fs[i][j](*([time] + x))]
                func_output += [np.stack(mini_out, axis=0)]

            return [np.concatenate(func_output, axis=1)] + self.step_states

        self.merged_func = merged_func
        self.merged_func.__name__ = utils.sanitize_name(
            "_".join([type(op.process).__name__ for op in self.ops])
        )

    def build_step(self, signals):
        time = [signals.gather(self.time_data)]
        input = [] if self.input_data is None else [signals.gather(self.input_data)]
        state = [signals.gather(s) for s in self.state_data]

        if compat.eager_enabled():
            # noop
            control_deps = contextlib.suppress()
        else:
            # we need to make sure that the previous call to this function
            # has completed before the next starts, since we don't know that the
            # functions are thread safe
            control_deps = tf.control_dependencies(self.prev_result)

        with control_deps:
            result = tf.numpy_function(
                self.merged_func,
                time + input + state,
                [self.output_data.dtype] + [s.dtype for s in self.state_data],
                name=self.merged_func.__name__,
            )

        # TensorFlow will automatically squeeze length-1 outputs (if there is
        # no state), which we don't want
        result = tf.nest.flatten(result)

        output = result[0]
        state = result[1:]

        self.prev_result = [output]

        output.set_shape(self.output_data.full_shape)
        signals.scatter(self.output_data, output, mode=self.mode)
        for i, s in enumerate(state):
            s.set_shape(self.state_data[i].full_shape)
            signals.scatter(self.state_data[i], s, mode="update")

    def build_post(self, signals):
        # generate state for each op
        step_states = [
            op.process.make_state(
                op.input.shape if op.input is not None else (0,),
                op.output.shape,
                signals.dt_val,
            )
            for op in self.ops
        ]

        # build all the states into combined array with shape
        # (n_states, n_ops, *state_d)
        combined_states = [
            [None for _ in self.ops] for _ in range(len(self.ops[0].state))
        ]
        for i, op in enumerate(self.ops):
            # note: we iterate over op.state so that the order is always based on that
            # dict's order (which is what we used to set up self.state_data)
            for j, name in enumerate(op.state):
                combined_states[j][i] = step_states[i][name]

        # combine op states, giving shape
        # (n_states, n_ops * state_d[0], *state_d[1:])
        # (keeping track of the offset of where each op's state lies in the
        # combined array)
        offsets = [[s.shape[0] for s in state] for state in combined_states]
        offsets = np.cumsum(offsets, axis=-1)
        self.step_states = [np.concatenate(state, axis=0) for state in combined_states]

        # cast to appropriate dtype
        for i, state in enumerate(self.state_data):
            self.step_states[i] = self.step_states[i].astype(state.dtype)

        # duplicate state for each minibatch, giving shape
        # (n_states, minibatch_size, n_ops * state_d[0], *state_d[1:])
        assert all(s.minibatched for op in self.ops for s in op.state.values())
        for i, state in enumerate(self.step_states):
            self.step_states[i] = np.tile(
                state[None, ...], (signals.minibatch_size,) + (1,) * state.ndim
            )

        # build the step functions
        self.step_fs = [[None for _ in range(signals.minibatch_size)] for _ in self.ops]
        for i, op in enumerate(self.ops):
            for j in range(signals.minibatch_size):
                # pass each make_step function a view into the combined state
                state = {}
                for k, name in enumerate(op.state):
                    start = 0 if i == 0 else offsets[k][i - 1]
                    stop = offsets[k][i]

                    state[name] = self.step_states[k][j, start:stop]

                    assert np.allclose(state[name], step_states[i][name])

                self.step_fs[i][j] = op.process.make_step(
                    op.input.shape if op.input is not None else (0,),
                    op.output.shape,
                    signals.dt_val,
                    op.process.get_rng(self.config.rng),
                    state,
                )

                self.step_fs[i][j] = utils.align_func(self.output_data.dtype)(
                    self.step_fs[i][j]
                )


class LowpassBuilder(OpBuilder):
    """Build a group of `~nengo.Lowpass` synapse operators."""

    def build_pre(self, signals, config):
        super().build_pre(signals, config)

        # the main difference between this and the general linearfilter
        # OneX implementation is that this version allows us to merge
        # synapses with different input dimensionality (by duplicating
        # the synapse parameters for every input, rather than using
        # broadcasting)

        self.input_data = signals.combine([op.input for op in self.ops])
        self.output_data = signals.combine([op.output for op in self.ops])

        nums = []
        dens = []
        for op in self.ops:
            if op.process.tau <= 0.03 * signals.dt_val:
                num = 1
                den = 0
            else:
                num, den, _ = cont2discrete(
                    (op.process.num, op.process.den), signals.dt_val, method="zoh"
                )
                num = num.flatten()

                num = num[1:] if num[0] == 0 else num
                assert len(num) == 1
                num = num[0]

                assert len(den) == 2
                den = den[1]

            nums += [num] * op.input.shape[0]
            dens += [den] * op.input.shape[0]

        if self.input_data.minibatched:
            # add batch dimension for broadcasting
            nums = np.expand_dims(nums, 0)
            dens = np.expand_dims(dens, 0)

        # apply the negative here
        dens = -np.asarray(dens)

        self.nums = tf.constant(nums, dtype=self.output_data.dtype)
        self.dens = tf.constant(dens, dtype=self.output_data.dtype)

        # create a variable to represent the internal state of the filter
        # self.state_sig = signals.make_internal(
        #     "state", self.output_data.shape)

    def build_step(self, signals):
        input = signals.gather(self.input_data)
        output = signals.gather(self.output_data)

        signals.scatter(self.output_data, self.dens * output + self.nums * input)

        # method using internal state signal
        # note: this isn't used for efficiency reasons (we can avoid an extra
        # scatter by reusing the output signal as the state signal)
        # input = signals.gather(self.input_data)
        # prev_state = signals.gather(self.state_sig)
        # new_state = self.dens * prev_state + self.nums * input
        # signals.scatter(self.state_sig, new_state)
        # signals.scatter(self.output_data, new_state)


class LinearFilterBuilder(OpBuilder):
    """Build a group of `~nengo.LinearFilter` synapse operators."""

    def build_pre(self, signals, config):
        super().build_pre(signals, config)

        # note: linear filters are linear systems with n_inputs/n_outputs == 1.
        # we apply them to multidimensional inputs, but we do so by
        # broadcasting that SISO linear system (so it's effectively
        # d 1-dimensional linear systems). this means that we can make
        # some simplifying assumptions, namely that B has shape (state_d, 1),
        # C has shape (1, state_d), and D has shape (1, 1), and then we can
        # implement those operations as (broadcasted) multiplies rather than
        # full matrix multiplications.
        # this also means that the minibatch dimension is identical to the
        # signal dimension (i.e. n m-dimensional signals is the same as
        # 1 n*m-dimensional signal); in either case we're just doing that
        # B/C/D broadcasting along all the non-state dimensions. so in these
        # computations we collapse minibatch and signal dimensions into one.

        self.input_data = signals.combine([op.input for op in self.ops])
        self.output_data = signals.combine([op.output for op in self.ops])

        if self.input_data.ndim != 1:
            raise NotImplementedError(
                "LinearFilter of non-vector signals is not implemented"
            )

        steps = [
            op.process.make_step(
                op.input.shape,
                op.output.shape,
                signals.dt_val,
                state=op.process.make_state(
                    op.input.shape, op.output.shape, signals.dt_val
                ),
                rng=None,
            )
            for op in self.ops
        ]
        self.filter_step = steps[0]
        assert all(isinstance(step, type(self.filter_step)) for step in steps)

        self.n_ops = len(self.ops)
        self.signal_d = self.ops[0].input.shape[0]
        self.state_d = steps[0].A.shape[0]

        if isinstance(self.filter_step, LinearFilter.NoX):
            self.A = None
            self.B = None
            self.C = None
            # combine D scalars for each op, and broadcast along minibatch and
            # signal dimensions
            self.D = tf.constant(
                np.concatenate([step.D[None, :, None] for step in steps], axis=1),
                dtype=signals.dtype,
            )

            assert self.D.shape == (1, self.n_ops, 1)
        elif isinstance(self.filter_step, LinearFilter.OneX):
            # combine A scalars for each op, and broadcast along batch/state
            self.A = tf.constant(
                np.concatenate([step.A for step in steps])[None, :], dtype=signals.dtype
            )
            # combine B and C scalars for each op, and broadcast along batch/state
            self.B = tf.constant(
                np.concatenate([step.B * step.C for step in steps])[None, :],
                dtype=signals.dtype,
            )
            self.C = None
            self.D = None

            assert self.A.shape == (1, self.n_ops, 1)
            assert self.B.shape == (1, self.n_ops, 1)
        else:
            self.A = tf.constant(
                np.stack([step.A for step in steps], axis=0), dtype=signals.dtype
            )
            self.B = tf.constant(
                np.stack([step.B for step in steps], axis=0), dtype=signals.dtype
            )
            self.C = tf.constant(
                np.stack([step.C for step in steps], axis=0), dtype=signals.dtype
            )

            if isinstance(self.filter_step, LinearFilter.NoD):
                self.D = None
            else:
                self.D = tf.constant(
                    np.concatenate([step.D[:, None, None] for step in steps]),
                    dtype=signals.dtype,
                )
                assert self.D.shape == (self.n_ops, 1, 1)

            # create a variable to represent the internal state of the filter
            self.state_data = signals.combine([op.state["X"] for op in self.ops])

            assert self.A.shape == (self.n_ops, self.state_d, self.state_d)
            assert self.B.shape == (self.n_ops, self.state_d, 1)
            assert self.C.shape == (self.n_ops, 1, self.state_d)

    def build_step(self, signals):
        input = signals.gather(self.input_data)

        if isinstance(self.filter_step, LinearFilter.NoX):
            input = tf.reshape(input, (signals.minibatch_size, self.n_ops, -1))

            signals.scatter(self.output_data, self.D * input)
        elif isinstance(self.filter_step, LinearFilter.OneX):
            input = tf.reshape(input, (signals.minibatch_size, self.n_ops, -1))

            # note: we use the output signal in place of a separate state
            output = signals.gather(self.output_data)
            output = tf.reshape(output, (signals.minibatch_size, self.n_ops, -1))

            signals.scatter(self.output_data, self.A * output + self.B * input)
        else:
            # TODO: possible to rework things to not require all the
            #  transposing/reshaping required for moving batch to end?
            def undo_batch(x):
                x = tf.reshape(x, x.shape[:-1].as_list() + [-1, signals.minibatch_size])
                x = tf.transpose(x, np.roll(np.arange(x.shape.ndims), 1))
                return x

            # separate by op and collapse batch/state dimensions
            assert input.shape.ndims == 2
            input = tf.transpose(input)
            input = tf.reshape(
                input, (self.n_ops, 1, self.signal_d * signals.minibatch_size)
            )

            state = signals.gather(self.state_data)
            assert input.shape.ndims == 3
            state = tf.transpose(state, perm=(1, 2, 0))
            state = tf.reshape(
                state,
                (self.n_ops, self.state_d, self.signal_d * signals.minibatch_size),
            )

            if isinstance(self.filter_step, LinearFilter.NoD):
                # for NoD, we update the state before computing the output
                new_state = tf.matmul(self.A, state) + self.B * input

                signals.scatter(self.state_data, undo_batch(new_state))

                output = tf.matmul(self.C, new_state)

                signals.scatter(self.output_data, undo_batch(output))
            else:
                # in the general case, we compute the output before updating
                # the state
                output = tf.matmul(self.C, state)
                if isinstance(self.filter_step, LinearFilter.General):
                    output += self.D * input
                signals.scatter(self.output_data, undo_batch(output))

                new_state = tf.matmul(self.A, state) + self.B * input

                signals.scatter(self.state_data, undo_batch(new_state))


@Builder.register(SimProcess)
class SimProcessBuilder(OpBuilder):
    """
    Builds a group of `~nengo.builder.processes.SimProcess` operators.

    Calls the appropriate sub-build class for the different process types.

    Attributes
    ----------
    TF_PROCESS_IMPL : dict of {`~nengo.Process`: `.builder.OpBuilder`}
        Mapping from process types to custom build classes (processes without
        a custom builder will use the generic builder).
    """

    # it is important that Lowpass come before LinearFilter because we'll be using
    # isinstance to find the right builder, and Lowpass is a subclass of LinearFilter
    TF_PROCESS_IMPL = {Lowpass: LowpassBuilder, LinearFilter: LinearFilterBuilder}

    def __init__(self, ops):
        super().__init__(ops)

        logger.debug("process %s", [op.process for op in ops])
        logger.debug("input %s", [op.input for op in ops])
        logger.debug("output %s", [op.output for op in ops])
        logger.debug("t %s", [op.t for op in ops])

        # if we have a custom tensorflow implementation for this process type,
        # then we build that. otherwise we'll execute the process step
        # function externally (using `tf.py_func`).
        for process_type, process_builder in self.TF_PROCESS_IMPL.items():
            if isinstance(ops[0].process, process_type):
                self.built_process = process_builder(ops)
                break
        else:
            self.built_process = GenericProcessBuilder(ops)

    def build_pre(self, signals, config):
        self.built_process.build_pre(signals, config)

    def build_step(self, signals):
        self.built_process.build_step(signals)

    def build_post(self, signals):
        if isinstance(self.built_process, GenericProcessBuilder):
            self.built_process.build_post(signals)

    @staticmethod
    def mergeable(x, y):
        # we can merge ops if they have a custom implementation, or merge
        # generic processes, but can't mix the two
        custom_impl = tuple(SimProcessBuilder.TF_PROCESS_IMPL.keys())
        if isinstance(x.process, custom_impl):
            if type(x.process) == Lowpass or type(y.process) == Lowpass:
                # lowpass ops can only be merged with other lowpass ops, since
                # they have a custom implementation
                if type(x.process) != type(y.process):  # noqa: E721
                    return False
            elif isinstance(x.process, LinearFilter):
                # we can only merge linearfilters that have the same state
                # dimensionality (den), the same step type (num), and the same
                # input signal dimensionality
                if (
                    not isinstance(y.process, LinearFilter)
                    or len(y.process.num) != len(x.process.num)
                    or len(y.process.den) != len(x.process.den)
                    or x.input.shape[0] != y.input.shape[0]
                ):
                    return False
            else:
                raise NotImplementedError()
        elif isinstance(y.process, custom_impl):
            return False

        return True



# pylint: disable=wrong-import-order,wrong-import-position,missing-docstring,ungrouped-imports
# isort: skip_file

__license__ = "Free for non-commercial use; see LICENSE.rst"
from nengo_dl.version import version as __version__
from nengo_dl.version import copyright as __copyright__

# filter out "INFO" level log messages
import os

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"

del os

# need to explicitly import these to trigger the builder registration
from nengo_dl import (
    op_builders,
    neuron_builders,
    process_builders,
    learning_rule_builders,
    transform_builders,
)

# import into top-level namespace
from nengo_dl import callbacks, compat, converter, dists, losses
from nengo_dl.config import configure_settings
from nengo_dl.converter import Converter
from nengo_dl.neurons import (
    LeakyReLU,
    SoftLIFRate,
    SpikingLeakyReLU,
)
from nengo_dl.simulator import Simulator
from nengo_dl.tensor_node import Layer, TensorNode, tensor_layer



# pylint: disable=missing-docstring

from nengo_dl import compat


def test_ordered_set():
    my_tuple = ("c", "b", "a")
    my_frozen_set = compat.FrozenOrderedSet(my_tuple)

    # order is preserved
    assert tuple(my_frozen_set) == my_tuple

    # test member functions
    for x in my_tuple:
        assert x in my_frozen_set

    assert "z" not in my_frozen_set

    assert len(my_frozen_set) == len(my_tuple)

    # hashable
    assert {my_frozen_set: "val"}[my_frozen_set] == "val"



"""
Build classes for Nengo neuron operators.
"""

import contextlib
import logging
import warnings

import numpy as np
import tensorflow as tf
from nengo.builder.neurons import SimNeurons
from nengo.neurons import LIF, LIFRate, RectifiedLinear, Sigmoid, SpikingRectifiedLinear
from tensorflow.python.framework import smart_cond

from nengo_dl import compat, utils
from nengo_dl.builder import Builder, OpBuilder
from nengo_dl.neurons import LeakyReLU, SoftLIFRate, SpikingLeakyReLU

logger = logging.getLogger(__name__)


class GenericNeuronBuilder(OpBuilder):
    """
    Builds all neuron types for which there is no custom TensorFlow
    implementation.

    Notes
    -----
    These will be executed as native Python functions, requiring execution to
    move in and out of TensorFlow.  This can significantly slow down the
    simulation, so any performance-critical neuron models should consider
    adding a custom TensorFlow implementation for their neuron type instead.
    """

    def build_pre(self, signals, config):
        super().build_pre(signals, config)

        self.J_data = signals.combine([op.J for op in self.ops])
        self.output_data = signals.combine([op.output for op in self.ops])

        state_keys = compat.neuron_state(self.ops[0]).keys()
        self.state_data = [
            signals.combine([compat.neuron_state(op)[key] for op in self.ops])
            for key in state_keys
        ]

        self.prev_result = []

        def neuron_step(dt, J, *states):  # pragma: no cover (runs in TF)
            output = None
            J_offset = 0
            state_offset = [0 for _ in states]
            for op in self.ops:
                # slice out the individual state vectors from the overall
                # array
                op_J = J[:, J_offset : J_offset + op.J.shape[0]]
                J_offset += op.J.shape[0]

                op_states = []
                for j, key in enumerate(state_keys):
                    s = compat.neuron_state(op)[key]
                    op_states += [
                        states[j][:, state_offset[j] : state_offset[j] + s.shape[0]]
                    ]
                    state_offset[j] += s.shape[0]

                # call neuron step function
                # note: `op_states` are views into `states`, which will
                # be updated in-place
                mini_out = []
                for j in range(signals.minibatch_size):
                    # blank output variable
                    neuron_output = np.zeros(op.output.shape, self.output_data.dtype)
                    compat.neuron_step(
                        op,
                        dt,
                        op_J[j],
                        neuron_output,
                        dict(zip(state_keys, [s[j] for s in op_states])),
                    )
                    mini_out.append(neuron_output)
                neuron_output = np.stack(mini_out, axis=0)

                # concatenate outputs
                if output is None:
                    output = neuron_output
                else:
                    output = np.concatenate((output, neuron_output), axis=1)

            return (output,) + states

        self.neuron_step = neuron_step
        self.neuron_step.__name__ = utils.sanitize_name(
            "_".join([repr(op.neurons) for op in self.ops])
        )

    def build_step(self, signals):
        J = signals.gather(self.J_data)
        states = [signals.gather(x) for x in self.state_data]
        states_dtype = [x.dtype for x in self.state_data]

        if compat.eager_enabled():
            # noop
            control_deps = contextlib.suppress()
        else:
            # we need to make sure that the previous call to this function
            # has completed before the next starts, since we don't know that the
            # functions are thread safe
            control_deps = tf.control_dependencies(self.prev_result)

        with control_deps:
            ret = tf.numpy_function(
                self.neuron_step,
                [signals.dt, J] + states,
                [self.output_data.dtype] + states_dtype,
                name=self.neuron_step.__name__,
            )

        neuron_out, state_out = ret[0], ret[1:]

        self.prev_result = [neuron_out]

        neuron_out.set_shape((signals.minibatch_size,) + self.output_data.shape)
        signals.scatter(self.output_data, neuron_out)

        for i, s in enumerate(self.state_data):
            state_out[i].set_shape((signals.minibatch_size,) + s.shape)
            signals.scatter(s, state_out[i])


class TFNeuronBuilder(OpBuilder):
    """Base class for `~nengo.neurons.NeuronType` builders with a TF implementation."""

    # TODO: this can be delegated to op.neurons.spiking if we increase the minimum
    #  Nengo version to one where that attribute is guaranteed to exist
    spiking = False

    def build_pre(self, signals, config):
        super().build_pre(signals, config)

        if hasattr(self.ops[0].neurons, "amplitude"):
            if all(op.neurons.amplitude == 1 for op in self.ops):
                self.amplitude = None
            else:
                self.amplitude = signals.op_constant(
                    [op.neurons for op in self.ops],
                    [op.J.shape[0] for op in self.ops],
                    "amplitude",
                    signals.dtype,
                )
        else:
            self.amplitude = None
        self.alpha = 1 if self.amplitude is None else self.amplitude
        self.alpha /= signals.dt

        self.J_data = signals.combine([op.J for op in self.ops])
        self.output_data = signals.combine([op.output for op in self.ops])

        self.state_data = {
            state: signals.combine([compat.neuron_state(op)[state] for op in self.ops])
            for state in compat.neuron_state(self.ops[0])
        }

    def step(self, J, dt, **state):
        """
        Implements the logic for a single inference step.

        If the neuron has no states, returns only the neuron output. Otherwise, returns
        a tuple where the first element is the neuron output, and subsequent elements
        correspond to each of the neuron's states. The order of the states must be the
        same as the order they appear in the neuron type's ``state`` dictionary.
        """
        raise NotImplementedError("Subclasses must implement")

    def training_step(self, J, dt, **state):
        """
        Implements the logic for a single training step.

        Returns only the neuron output. Therefore, the ``training_step`` cannot affect
        any neuron states; they will all be held constant during training.

        Note: subclasses only need to implement this if ``spiking=True``. It is used
        to specify an alternate (differentiable) implementation of the neuron model
        to be used during training.
        """

    def build_step(self, signals, **step_kwargs):
        J = signals.gather(self.J_data)
        state = {s: signals.gather(d) for s, d in self.state_data.items()}

        step_output = self.step(J, signals.dt, **state)
        if isinstance(step_output, tuple) and len(step_output) != 1 + len(state):
            raise ValueError(
                f"`{type(self).__name__}.step` must return a tuple with the neuron "
                f"output followed by tensors with each updated state (one tensor per "
                f"state). `{type(self).__name__}` has {len(state)} states, but only "
                f"received output for {len(step_output) - 1} states."
            )

        step_output = tf.nest.flatten(step_output)

        if not self.spiking or self.config.inference_only:
            out = step_output
        else:
            out = tf.nest.flatten(
                smart_cond.smart_cond(
                    self.config.training,
                    true_fn=lambda: (self.training_step(J, signals.dt, **state),)
                    + tuple(state.values()),
                    # we use stop_gradient to avoid propagating any nans (those get
                    # propagated through the cond even if the spiking version isn't
                    # being used at all)
                    false_fn=lambda: tuple(
                        tf.stop_gradient(x) for x in tf.nest.flatten(step_output)
                    ),
                )
            )

        signals.scatter(self.output_data, out[0])
        for state_data, v in zip(self.state_data.values(), out[1:]):
            signals.scatter(state_data, v)


class RectifiedLinearBuilder(TFNeuronBuilder):
    """Build a group of `~nengo.RectifiedLinear` neuron operators."""

    def build_pre(self, signals, config):
        super().build_pre(signals, config)

        if all(getattr(op.neurons, "negative_slope", 0) == 0 for op in self.ops):
            self.negative_slope = None
        else:
            self.negative_slope = signals.op_constant(
                [op.neurons for op in self.ops],
                [op.J.shape[0] for op in self.ops],
                "negative_slope",
                signals.dtype,
            )

    def step(self, J, dt):
        out = tf.nn.relu(J)
        if self.negative_slope is not None:
            out -= self.negative_slope * tf.nn.relu(-J)
        if self.amplitude is not None:
            out *= self.amplitude
        return out


class SpikingRectifiedLinearBuilder(RectifiedLinearBuilder):
    """Build a group of `~nengo.SpikingRectifiedLinear` neuron operators."""

    spiking = True

    def step(self, J, dt, voltage):
        if self.negative_slope is None:
            voltage += tf.nn.relu(J) * dt
            n_spikes = tf.floor(voltage)
        else:
            voltage += (tf.nn.relu(J) - self.negative_slope * tf.nn.relu(-J)) * dt
            n_spikes = tf.floor(voltage) + tf.cast(voltage < 0, voltage.dtype)

        voltage -= n_spikes
        out = n_spikes * self.alpha

        return out, voltage

    def training_step(self, J, dt, **state):
        return super().step(J, dt)


class SigmoidBuilder(TFNeuronBuilder):
    """Build a group of `~nengo.Sigmoid` neuron operators."""

    def build_pre(self, signals, config):
        super().build_pre(signals, config)

        self.tau_ref = signals.op_constant(
            [op.neurons for op in self.ops],
            [op.J.shape[0] for op in self.ops],
            "tau_ref",
            signals.dtype,
        )

    def step(self, J, dt):
        return tf.nn.sigmoid(J) / self.tau_ref


class TanhBuilder(TFNeuronBuilder):
    """Build a group of `~nengo.Tanh` neuron operators."""

    def build_pre(self, signals, config):
        super().build_pre(signals, config)

        self.tau_ref = signals.op_constant(
            [op.neurons for op in self.ops],
            [op.J.shape[0] for op in self.ops],
            "tau_ref",
            signals.dtype,
        )

    def step(self, J, dt):
        return tf.nn.tanh(J) / self.tau_ref


class LIFRateBuilder(TFNeuronBuilder):
    """Build a group of `~nengo.LIFRate` neuron operators."""

    def build_pre(self, signals, config):
        super().build_pre(signals, config)

        self.tau_ref = signals.op_constant(
            [op.neurons for op in self.ops],
            [op.J.shape[0] for op in self.ops],
            "tau_ref",
            signals.dtype,
        )
        self.tau_rc = signals.op_constant(
            [op.neurons for op in self.ops],
            [op.J.shape[0] for op in self.ops],
            "tau_rc",
            signals.dtype,
        )

        self.zeros = tf.zeros(
            (signals.minibatch_size,) + self.J_data.shape, signals.dtype
        )

        self.epsilon = tf.constant(1e-15, dtype=signals.dtype)

        # copy these so that they're easily accessible in the step functions
        self.zero = signals.zero
        self.one = signals.one

    def step(self, J, dt):
        J -= self.one

        # note: we convert all the j to be positive before this calculation
        # (even though we'll only use the values that are already positive),
        # otherwise we can end up with nans in the gradient
        rates = (self.one if self.amplitude is None else self.amplitude) / (
            self.tau_ref
            + self.tau_rc
            * tf.math.log1p(tf.math.reciprocal(tf.maximum(J, self.epsilon)))
        )

        return tf.where(J > self.zero, rates, self.zeros)


class SoftLIFRateBuilder(LIFRateBuilder):
    """Build a group of `.SoftLIFRate` neuron operators."""

    def build_pre(self, signals, config):
        super().build_pre(signals, config)

        self.sigma = signals.op_constant(
            [op.neurons for op in self.ops],
            [op.J.shape[0] for op in self.ops],
            "sigma",
            signals.dtype,
        )

    def step(self, J, dt):
        J -= self.one

        js = J / self.sigma
        j_valid = js > -20
        js_safe = tf.where(j_valid, js, self.zeros)

        # softplus(js) = log(1 + e^js)
        z = tf.nn.softplus(js_safe) * self.sigma

        # as z->0
        #   z = s*log(1 + e^js) = s*e^js
        #   log(1 + 1/z) = log(1/z) = -log(s*e^js) = -js - log(s)
        q = tf.where(
            j_valid, tf.math.log1p(tf.math.reciprocal(z)), -js - tf.math.log(self.sigma)
        )

        rates = (self.one if self.amplitude is None else self.amplitude) / (
            self.tau_ref + self.tau_rc * q
        )

        return rates


class LIFBuilder(SoftLIFRateBuilder):
    """Build a group of `~nengo.LIF` neuron operators."""

    spiking = True

    def build_pre(self, signals, config):
        # note: we skip the SoftLIFRateBuilder init
        # pylint: disable=bad-super-call
        super(SoftLIFRateBuilder, self).build_pre(signals, config)

        self.min_voltage = signals.op_constant(
            [op.neurons for op in self.ops],
            [op.J.shape[0] for op in self.ops],
            "min_voltage",
            signals.dtype,
        )

        if self.config.lif_smoothing:
            self.sigma = tf.constant(self.config.lif_smoothing, dtype=signals.dtype)

    def step(self, J, dt, voltage, refractory_time):
        delta_t = tf.clip_by_value(dt - refractory_time, self.zero, dt)

        dV = (voltage - J) * tf.math.expm1(
            -delta_t / self.tau_rc  # pylint: disable=invalid-unary-operand-type
        )
        voltage += dV

        spiked = voltage > self.one
        spikes = tf.cast(spiked, J.dtype) * self.alpha

        partial_ref = -self.tau_rc * tf.math.log1p(
            (self.one - voltage) / (J - self.one)
        )
        # FastLIF version (linearly approximate spike time when calculating
        # remaining refractory period)
        # partial_ref = signals.dt * (voltage - self.one) / dV

        refractory_time = tf.where(
            spiked, self.tau_ref - partial_ref, refractory_time - dt
        )

        voltage = tf.where(spiked, self.zeros, tf.maximum(voltage, self.min_voltage))

        return spikes, voltage, refractory_time

    def training_step(self, J, dt, **state):
        return (
            LIFRateBuilder.step(self, J, dt)
            if self.config.lif_smoothing is None
            else SoftLIFRateBuilder.step(self, J, dt)
        )


class RegularSpikingBuilder(TFNeuronBuilder):
    """Build a group of `~nengo.RegularSpiking` neuron operators."""

    spiking = True

    def step(self, J, dt, voltage):
        voltage += J * dt
        n_spikes = tf.floor(voltage)

        voltage -= n_spikes
        out = n_spikes * self.alpha

        return out, voltage

    def training_step(self, J, dt, **state):
        return J if self.amplitude is None else J * self.amplitude


class StochasticSpikingBuilder(TFNeuronBuilder):
    """Build a group of `~nengo.StochasticSpiking` neuron operators."""

    spiking = True

    def step(self, J, dt):
        x = dt * tf.math.abs(J)
        n_spikes = tf.floor(x)
        frac = x - n_spikes

        n_spikes += tf.cast(
            tf.random.uniform(frac.shape, dtype=frac.dtype) < frac, n_spikes.dtype
        )

        n_spikes *= self.alpha * tf.math.sign(J)

        return n_spikes

    def training_step(self, J, dt):
        return J if self.amplitude is None else J * self.amplitude


class PoissonSpikingBuilder(TFNeuronBuilder):
    """Build a group of `~nengo.PoissonSpiking` neuron operators."""

    spiking = True

    def step(self, J, dt):
        n_spikes = (
            self.alpha
            * tf.random.poisson((), tf.math.abs(J) * dt, dtype=J.dtype)
            * tf.math.sign(J)
        )
        n_spikes.set_shape(J.shape)

        return n_spikes

    def training_step(self, J, dt):
        return J if self.amplitude is None else J * self.amplitude


@Builder.register(SimNeurons)
class SimNeuronsBuilder(OpBuilder):
    """
    Builds a group of `~nengo.builder.neurons.SimNeurons` operators.

    Calls the appropriate sub-build class for the different neuron types.

    Attributes
    ----------
    TF_NEURON_IMPL : dict of {`~nengo.neurons.NeuronType`, \
                              `.builder.OpBuilder`}
        Mapping from neuron types to custom build classes (neurons without
        a custom builder will use the generic builder).
    """

    TF_NEURON_IMPL = {
        RectifiedLinear: RectifiedLinearBuilder,
        SpikingRectifiedLinear: SpikingRectifiedLinearBuilder,
        LeakyReLU: RectifiedLinearBuilder,
        SpikingLeakyReLU: SpikingRectifiedLinearBuilder,
        Sigmoid: SigmoidBuilder,
        compat.Tanh: TanhBuilder,
        LIF: LIFBuilder,
        LIFRate: LIFRateBuilder,
        SoftLIFRate: SoftLIFRateBuilder,
        compat.RegularSpiking: RegularSpikingBuilder,
        compat.StochasticSpiking: StochasticSpikingBuilder,
        compat.PoissonSpiking: PoissonSpikingBuilder,
    }

    def __init__(self, ops):
        super().__init__(ops)

        neuron_type = type(ops[0].neurons)

        # if we have a custom tensorflow implementation for this neuron type,
        # then we build that. otherwise we'll just execute the neuron step
        # function externally (using `tf.py_func`).
        if neuron_type in self.TF_NEURON_IMPL:
            self.built_neurons = self.TF_NEURON_IMPL[neuron_type](ops)
        else:
            warnings.warn(
                f"{neuron_type} does not have a native TensorFlow implementation; "
                "falling back to Python implementation"
            )
            self.built_neurons = GenericNeuronBuilder(ops)

    def build_pre(self, signals, config):
        self.built_neurons.build_pre(signals, config)

    def build_step(self, signals):
        self.built_neurons.build_step(signals)

    @staticmethod
    def mergeable(x, y):
        # neuron ops must all have the same type
        return type(x.neurons) == type(y.neurons)  # noqa: E721



# pylint: disable=unused-import,ungrouped-imports

"""
Utilities to ease cross-compatibility between different versions of upstream
dependencies.
"""

import collections
import inspect

import nengo
import tensorflow as tf
from nengo._vendor.scipy.sparse import linalg_interface, linalg_onenormest
from packaging import version
from tensorflow.python.eager import context

# TensorFlow compatibility


class TFLogFilter:
    """
    Filter for cleaning up the TensorFlow log output.

    Parameters
    ----------
    err_on_deprecation : bool
        If True, treat deprecation log messages as errors.
    """

    def __init__(self, err_on_deprecation):
        self.err_on_deprecation = err_on_deprecation

    def filter(self, record):
        """
        Apply filter to log record.

        Parameters
        ----------
        record : `LogRecord`
            Message emitted by logger.

        Returns
        -------
        pass : bool
            If True, pass the ``record``, otherwise filter it out.

        Raises
        ------
        AttributeError
            If a deprecation message is detected and ``err_on_deprecation=True``.
        """

        # "tf.keras.backend.get_session is deprecated": this deprecation message
        # is raised incorrectly due to a bug, see
        # https://github.com/tensorflow/tensorflow/issues/33182
        if len(record.args) > 1 and record.args[1] == "tf.keras.backend.get_session":
            return False

        # "Output steps_run missing from loss dictionary": steps_run should
        # never have a loss defined
        if record.msg.startswith("Output steps_run missing from loss dictionary"):
            return False

        if self.err_on_deprecation and (
            "deprecation.py" in record.pathname or "deprecated" in record.msg.lower()
        ):
            msg = record.getMessage()
            raise AttributeError(f"Deprecation warning detected:\n{msg}")

        return True


tf.get_logger().addFilter(TFLogFilter(err_on_deprecation=False))

if version.parse(tf.__version__) < version.parse("2.3.0rc0"):

    def _build_map(outputs):
        """
        Vendored from ``tensorflow.python.keras.engine.functional._build_map``
        in TF 2.3.0 (with a few changes noted below).
        """
        finished_nodes = set()
        nodes_in_progress = set()
        nodes_in_decreasing_depth = []  # nodes from inputs -> outputs.
        layer_indices = {}  # layer -> in traversal order.
        for output in tf.nest.flatten(outputs):
            _build_map_helper(
                output,
                finished_nodes,
                nodes_in_progress,
                nodes_in_decreasing_depth,
                layer_indices,
            )

        # CHANGED: add `node.layer` alias for `node.outbound_layer`
        for node in nodes_in_decreasing_depth:
            node.layer = node.outbound_layer

        return nodes_in_decreasing_depth, layer_indices

    def _build_map_helper(
        tensor,
        finished_nodes,
        nodes_in_progress,
        nodes_in_decreasing_depth,
        layer_indices,
    ):
        """Recursive helper for `_build_map`."""
        layer, node_index, _ = tensor._keras_history  # pylint: disable=protected-access
        node = layer._inbound_nodes[node_index]  # pylint: disable=protected-access

        # Don't repeat work for shared subgraphs
        if node in finished_nodes:
            return

        # Prevent cycles.
        if node in nodes_in_progress:  # pragma: no cover
            raise ValueError(
                "The tensor "
                + str(tensor)
                + ' at layer "'
                + layer.name
                + '" is part of a cycle.'
            )

        # Store the traversal order for layer sorting.
        if layer not in layer_indices:
            layer_indices[layer] = len(layer_indices)

        # Propagate to all previous tensors connected to this node.
        nodes_in_progress.add(node)
        # CHANGED: `not node.is_input` to `node.inbound_layers`
        if node.inbound_layers:
            # CHANGED: `node.keras_inputs` to `tf.nest.flatten(node.input_tensors)`
            # CHANGED: `tensor` to `input_tensor` (for pylint)
            for input_tensor in tf.nest.flatten(node.input_tensors):
                _build_map_helper(
                    input_tensor,
                    finished_nodes,
                    nodes_in_progress,
                    nodes_in_decreasing_depth,
                    layer_indices,
                )

        finished_nodes.add(node)
        nodes_in_progress.remove(node)
        nodes_in_decreasing_depth.append(node)

    from tensorflow.keras import Model as Functional
    from tensorflow.python.keras.layers import (
        BatchNormalizationV1,
        BatchNormalizationV2,
    )
elif version.parse(tf.__version__) < version.parse("2.6.0rc0"):  # pragma: no cover
    from tensorflow.python.keras.engine.functional import Functional, _build_map
    from tensorflow.python.keras.layers import (
        BatchNormalizationV1,
        BatchNormalizationV2,
    )
else:
    from keras.engine.functional import Functional, _build_map
    from keras.layers import BatchNormalizationV1, BatchNormalizationV2


if version.parse(tf.__version__) < version.parse("2.3.0rc0"):
    from tensorflow.python.keras.engine import network

    # monkeypatch to fix bug in TF2.2, see
    # https://github.com/tensorflow/tensorflow/issues/37548
    old_conform = network.Network._conform_to_reference_input

    def _conform_to_reference_input(self, tensor, ref_input):
        keras_history = getattr(tensor, "_keras_history", None)

        tensor = old_conform(self, tensor, ref_input)

        if keras_history is not None:
            tensor._keras_history = keras_history

        return tensor

    network.Network._conform_to_reference_input = _conform_to_reference_input

if version.parse(tf.__version__) < version.parse("2.5.0rc0"):

    def sub_layers(layer):
        """Get layers contained in ``layer``."""
        return layer._layers


else:

    def sub_layers(layer):
        """Get layers contained in ``layer``."""
        return layer._self_tracked_trackables

    # monkeypatch to fix bug when using TF2.5 with sphinx's doctest extension
    from tensorflow.python.autograph.impl.api import StackTraceMapper

    old_source_map = StackTraceMapper.get_effective_source_map

    def get_effective_source_map(self):
        """
        Sometimes the source file is unknown (e.g. when running code through Sphinx's
        doctest builder). This causes TensorFlow to crash (as of TF 2.5). So we convert
        any Nones to the string "unknown".
        """

        effective_source_map = old_source_map(self)

        # convert Nones to "unknown"
        effective_source_map = {
            key: tuple("unknown" if x is None else x for x in val)
            for key, val in effective_source_map.items()
        }
        return effective_source_map

    StackTraceMapper.get_effective_source_map = get_effective_source_map


# Nengo compatibility

if version.parse(nengo.__version__) < version.parse("3.1.0"):
    PoissonSpiking = RegularSpiking = StochasticSpiking = Tanh = NoTransform = type(
        None
    )

    default_transform = 1

    def conn_has_weights(conn):
        """All connections have weights."""
        return True

    def neuron_state(neuron_op):
        """Look up keys from function signature."""
        names = list(inspect.signature(neuron_op.neurons.step_math).parameters.keys())[
            3:
        ]
        assert len(names) == len(neuron_op.states)
        return dict(zip(names, neuron_op.states))

    def neuron_step(neuron_op, dt, J, output, state):  # pragma: no cover (runs in TF)
        """Call step_math instead of step."""
        neuron_op.neurons.step_math(dt, J, output, *state.values())

    def to_neurons(conn):
        """Check whether the output of a connection is a neuron object."""
        return isinstance(conn.post_obj, nengo.ensemble.Neurons) or (
            isinstance(conn.pre_obj, nengo.Ensemble)
            and isinstance(conn.post_obj, nengo.Ensemble)
            and conn.solver.weights
        )

    class SimProbe(nengo.builder.Operator):
        """Backport of Nengo 3.1.0 SimProbe."""

        def __init__(self, signal, tag=None):
            super().__init__(tag=tag)
            self.sets = []
            self.incs = []
            self.reads = [signal]
            self.updates = []

        def make_step(self, signals, dt, rng):
            """Not used in NengoDL."""
            raise NotImplementedError

    from nengo_dl.builder import NengoBuilder as _NengoBuilder

    def build_probe(model, probe):
        """Copy of the base probe builder that adds the SimProbe op."""
        nengo.builder.probe.build_probe(model, probe)
        model.add_op(SimProbe(model.sig[probe]["in"]))

    _NengoBuilder.register(nengo.Probe)(build_probe)

    # monkeypatch fix for https://github.com/nengo/nengo/pull/1587
    linalg_onenormest.aslinearoperator = linalg_interface.aslinearoperator

else:
    from nengo.builder.probe import SimProbe
    from nengo.neurons import PoissonSpiking, RegularSpiking, StochasticSpiking, Tanh
    from nengo.transforms import NoTransform

    default_transform = None

    def conn_has_weights(conn):
        """Equivalent to conn.has_weights."""
        return conn.has_weights

    def neuron_state(neuron_op):
        """Equivalent to neuron_op.state."""
        return neuron_op.state

    def neuron_step(neuron_op, dt, J, output, state):  # pragma: no cover (runs in TF)
        """Equivalent to neuron_op.step."""
        neuron_op.neurons.step(dt, J, output, **state)

    def to_neurons(conn):
        """Equivalent to conn._to_neurons."""
        return conn._to_neurons


if version.parse(nengo.__version__) <= version.parse("3.1.0"):

    class FrozenOrderedSet(collections.abc.Set):
        """Backport of `nengo.utils.stdlib.FrozenOrderedSet`."""

        def __init__(self, data):
            self.data = dict((d, None) for d in data)

        def __contains__(self, elem):
            return elem in self.data

        def __iter__(self):
            return iter(self.data)

        def __len__(self):
            return len(self.data)

        def __hash__(self):
            return self._hash()


else:
    from nengo.utils.stdlib import FrozenOrderedSet


def eager_enabled():
    """
    Check if we're in eager mode or graph mode.

    Note: this function differs from ``tf.executing_eagerly()`` in that it will still
    return ``True`` if we're inside a ``tf.function``. Essentially this checks whether
    the user has called ``tf.compat.v1.disable_eager_execution()`` or not.
    """

    return context.default_execution_mode == context.EAGER_MODE


try:
    from keras_spiking import Alpha, Lowpass, SpikingActivation
except ImportError:
    SpikingActivation = object()
    Lowpass = object()
    Alpha = object()



# Automatically generated by nengo-bones, do not edit this file directly

"""NengoDL version information.

We use semantic versioning (see http://semver.org/)
and conform to PEP440 (see https://www.python.org/dev/peps/pep-0440/).
'.dev0' will be added to the version unless the code base represents
a release version. Release versions are git tagged with the version.
"""

version_info = (3, 4, 2)

name = "nengo-dl"
dev = None

# use old string formatting, so that this can still run in Python <= 3.5
# (since this file is parsed in setup.py, before python_requires is applied)
version = "{v}{dev}".format(
    v=".".join(str(v) for v in version_info),
    dev=(".dev%d" % dev) if dev is not None else "",
)

copyright = "Copyright (c) 2015-2021 Applied Brain Research"

import warnings  # noqa: E402 pylint: disable=wrong-import-position

# check nengo version
try:
    import nengo.version
except ImportError:
    # nengo not installed, can't check version
    pass
else:
    minimum_nengo_version = (3, 0, 0)

    # for release versions of nengo-dl, this should be the latest released
    # nengo version. for dev versions of nengo-dl, this should be the current
    # nengo dev version.
    latest_nengo_version = (3, 1, 0)

    if nengo.version.version_info < minimum_nengo_version:  # pragma: no cover
        raise ValueError(
            (
                "NengoDL does not support Nengo version {nengo_version}. "
                "Upgrade with 'pip install --upgrade --no-deps nengo'."
            ).format(nengo_version=nengo.version.version)
        )
    elif nengo.version.version_info > latest_nengo_version:  # pragma: no cover
        warnstr = (
            "This version of NengoDL has not been tested with your Nengo version "
            "({nengo_version}). The latest fully supported version is {latest_version}."
        ).format(
            nengo_version=nengo.version.version,
            latest_version=".".join(str(x) for x in latest_nengo_version),
        )
        warnings.warn(warnstr)
